# Copyright 2016 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Handles all processes relating to notifications.

The :py:class:`MasakariManager` class is a
:py:class:`masakari.manager.Manager` that handles RPC calls relating to
notifications. It is responsible for processing notifications and executing
workflows.

"""
import traceback

from oslo_log import log as logging
import oslo_messaging as messaging
from oslo_service import periodic_task
from oslo_utils import timeutils

import masakari.conf
from masakari.engine import driver
from masakari.engine import instance_events as virt_events
from masakari.engine import rpcapi
from masakari.engine import utils as engine_utils
from masakari import exception
from masakari.i18n import _
from masakari import manager
from masakari import objects
from masakari.objects import fields
from masakari import utils

CONF = masakari.conf.CONF

LOG = logging.getLogger(__name__)


def update_host_method(context, host_name, reserved=False):
    reserved_host = objects.Host.get_by_name(context, host_name)
    reserved_host.reserved = reserved
    reserved_host.save()


class MasakariManager(manager.Manager):
    """Manages the running notifications"""
    RPC_API_VERSION = rpcapi.EngineAPI.RPC_API_VERSION
    target = messaging.Target(version=RPC_API_VERSION)

    def __init__(self, masakari_driver=None, *args, **kwargs):
        """Load configuration options"""
        LOG.debug("Initializing Masakari Manager.")
        super(MasakariManager, self).__init__(service_name="engine",
                                             *args, **kwargs)

        self.driver = driver.load_masakari_driver(masakari_driver)

    def _handle_notification_type_process(self, context, notification):
        notification_status = fields.NotificationStatus.FINISHED
        notification_event = notification.payload.get('event')
        process_name = notification.payload.get('process_name')
        exception_info = None

        if notification_event.upper() == 'STARTED':
            LOG.info("Notification type '%(type)s' received for host "
                     "'%(host_uuid)s': '%(process_name)s' has been "
                     "%(event)s.",
                     {'type': notification.type,
                      'host_uuid': notification.source_host_uuid,
                      'process_name': process_name,
                      'event': notification_event})
        elif notification_event.upper() == 'STOPPED':
            host_obj = objects.Host.get_by_uuid(
                context, notification.source_host_uuid)
            host_name = host_obj.name

            # Mark host on_maintenance mode as True
            update_data = {
                'on_maintenance': True,
            }
            host_obj.update(update_data)
            host_obj.save()

            try:
                self.driver.execute_process_failure(
                    context, process_name, host_name,
                    notification.notification_uuid)
            except exception.SkipProcessRecoveryException:
                notification_status = fields.NotificationStatus.FINISHED
            except (exception.MasakariException,
                    exception.ProcessRecoveryFailureException) as e:
                notification_status = fields.NotificationStatus.ERROR
                LOG.error("Failed to process notification '%(uuid)s'."
                          " Reason: %(error)s",
                          {"uuid": notification.notification_uuid,
                           "error": e.message})
                exception_info = e
        else:
            LOG.warning("Invalid event: %(event)s received for "
                        "notification type: %(notification_type)s",
                        {'event': notification_event,
                         'notification_type': notification.type})
            notification_status = fields.NotificationStatus.IGNORED

        if exception_info:
            tb = traceback.format_exc()
            engine_utils.notify_about_notification_update(context,
                notification,
                action=fields.EventNotificationAction.NOTIFICATION_PROCESS,
                phase=fields.EventNotificationPhase.ERROR,
                exception=str(exception_info),
                tb=tb)
        else:
            engine_utils.notify_about_notification_update(context,
                notification,
                action=fields.EventNotificationAction.NOTIFICATION_PROCESS,
                phase=fields.EventNotificationPhase.END)

        return notification_status

    def _handle_notification_type_instance(self, context, notification):
        if not virt_events.is_valid_event(notification.payload):
            LOG.info("Notification '%(uuid)s' received with payload "
                     "%(payload)s is ignored.",
                     {"uuid": notification.notification_uuid,
                      "payload": notification.payload})
            return fields.NotificationStatus.IGNORED

        notification_status = fields.NotificationStatus.FINISHED
        exception_info = None
        try:
            self.driver.execute_instance_failure(
                context, notification.payload.get('instance_uuid'),
                notification.notification_uuid)
        except exception.IgnoreInstanceRecoveryException as e:
            notification_status = fields.NotificationStatus.IGNORED
            exception_info = e
        except exception.SkipInstanceRecoveryException:
            notification_status = fields.NotificationStatus.FINISHED
        except (exception.MasakariException,
                exception.InstanceRecoveryFailureException) as e:
            notification_status = fields.NotificationStatus.ERROR
            LOG.error("Failed to process notification '%(uuid)s'."
                      " Reason: %(error)s",
                      {"uuid": notification.notification_uuid,
                       "error": e.message})
            exception_info = e

        if exception_info:
            tb = traceback.format_exc()
            engine_utils.notify_about_notification_update(context,
                notification,
                action=fields.EventNotificationAction.NOTIFICATION_PROCESS,
                phase=fields.EventNotificationPhase.ERROR,
                exception=str(exception_info),
                tb=tb)
        else:
            engine_utils.notify_about_notification_update(context,
                notification,
                action=fields.EventNotificationAction.NOTIFICATION_PROCESS,
                phase=fields.EventNotificationPhase.END)

        return notification_status

    def _handle_notification_type_host(self, context, notification):
        host_status = notification.payload.get('host_status')
        notification_status = fields.NotificationStatus.FINISHED
        notification_event = notification.payload.get('event')
        exception_info = None

        if host_status is None:
            LOG.warning("Notification '%(uuid)s' ignored as host_status is "
                        "not provided.",
                        {'uuid': notification.notification_uuid})
            notification_status = fields.NotificationStatus.IGNORED
        elif host_status.upper() != fields.HostStatusType.NORMAL:
            # NOTE(shilpasd): Avoid host recovery for host_status other than
            # 'NORMAL' otherwise it could lead to unsafe evacuation of
            # instances running on the failed source host.
            LOG.warning("Notification '%(uuid)s' ignored as host_status "
                        "is '%(host_status)s'",
                        {'uuid': notification.notification_uuid,
                         'host_status': host_status.upper()})
            notification_status = fields.NotificationStatus.IGNORED
        elif notification_event.upper() == 'STARTED':
            LOG.info("Notification type '%(type)s' received for host "
                     "'%(host_uuid)s' has been %(event)s.",
                     {'type': notification.type,
                      'host_uuid': notification.source_host_uuid,
                      'event': notification_event})
        elif notification_event.upper() == 'STOPPED':
            host_obj = objects.Host.get_by_uuid(
                context, notification.source_host_uuid)
            host_name = host_obj.name
            recovery_method = host_obj.failover_segment.recovery_method

            # Mark host on_maintenance mode as True
            update_data = {
                'on_maintenance': True,
            }

            # Set reserved flag to False if this host is reserved
            if host_obj.reserved:
                update_data['reserved'] = False

            LOG.info("Set host %s on maintenance.", host_name)
            host_obj.update(update_data)
            host_obj.save()

            reserved_host_list = None

            if not recovery_method == (
                    fields.FailoverSegmentRecoveryMethod.AUTO):
                reserved_host_object_list = objects.HostList.get_all(
                    context, filters={
                        'failover_segment_id': host_obj.failover_segment.uuid,
                        'reserved': True,
                        'on_maintenance': False
                        })
                # Create list of host name from reserved_host_object_list
                reserved_host_list = [host.name for host in
                                      reserved_host_object_list]

            try:
                self.driver.execute_host_failure(
                    context, host_name, recovery_method,
                    notification.notification_uuid,
                    update_host_method=update_host_method,
                    reserved_host_list=reserved_host_list)
            except exception.SkipHostRecoveryException:
                notification_status = fields.NotificationStatus.FINISHED
            except (exception.HostRecoveryFailureException,
                    exception.ReservedHostsUnavailable,
                    exception.MasakariException) as e:
                notification_status = fields.NotificationStatus.ERROR
                LOG.error("Failed to process notification '%(uuid)s'."
                          " Reason: %(error)s",
                          {"uuid": notification.notification_uuid,
                           "error": e.message})
                exception_info = e
        else:
            LOG.warning("Invalid event: %(event)s received for "
                        "notification type: %(type)s",
                        {'event': notification_event,
                         'type': notification.type})
            notification_status = fields.NotificationStatus.IGNORED

        if exception_info:
            tb = traceback.format_exc()
            engine_utils.notify_about_notification_update(context,
                notification,
                action=fields.EventNotificationAction.NOTIFICATION_PROCESS,
                phase=fields.EventNotificationPhase.ERROR,
                exception=str(exception_info),
                tb=tb)
        else:
            engine_utils.notify_about_notification_update(context,
                notification,
                action=fields.EventNotificationAction.NOTIFICATION_PROCESS,
                phase=fields.EventNotificationPhase.END)

        return notification_status

    def _process_notification(self, context, notification):
        @utils.synchronized(notification.source_host_uuid, blocking=True)
        def do_process_notification(notification):
            LOG.info('Processing notification %(notification_uuid)s of '
                     'type: %(type)s',
                     {'notification_uuid': notification.notification_uuid,
                      'type': notification.type})

            # Get notification from db
            notification_db = objects.Notification.get_by_uuid(context,
                                        notification.notification_uuid)

            # NOTE(tpatil): To fix bug 1773132, process notification only
            # if the notification status is New and the current notification
            # from DB status is Not New to avoid recovering from failure twice
            if (notification.status == fields.NotificationStatus.NEW and
                    notification_db.status != fields.NotificationStatus.NEW):
                LOG.warning("Processing of notification is skipped to avoid "
                            "recovering from failure twice. "
                            "Notification received is '%(uuid)s' "
                            "and it's status is '%(new_status)s' and the "
                            "current status of same notification in db "
                            "is '%(old_status)s'",
                            {"uuid": notification.notification_uuid,
                            "new_status": notification.status,
                            "old_status": notification_db.status})
                return

            update_data = {
                'status': fields.NotificationStatus.RUNNING,
            }
            notification.update(update_data)
            notification.save()

            if notification.type == fields.NotificationType.PROCESS:
                notification_status = self._handle_notification_type_process(
                    context, notification)
            elif notification.type == fields.NotificationType.VM:
                notification_status = self._handle_notification_type_instance(
                    context, notification)
            elif notification.type == fields.NotificationType.COMPUTE_HOST:
                notification_status = self._handle_notification_type_host(
                    context, notification)

            LOG.info("Notification %(notification_uuid)s exits with "
                     "status: %(status)s.",
                     {'notification_uuid': notification.notification_uuid,
                      'status': notification_status})

            update_data = {
                'status': notification_status
            }
            notification.update(update_data)
            notification.save()

        engine_utils.notify_about_notification_update(context,
            notification,
            action=fields.EventNotificationAction.NOTIFICATION_PROCESS,
            phase=fields.EventNotificationPhase.START)

        do_process_notification(notification)

    def process_notification(self, context, notification=None):
        """Processes the notification"""
        host = objects.Host.get_by_uuid(
            context, notification.source_host_uuid)
        if not host.failover_segment.enabled:
            update_data = {
                'status': fields.NotificationStatus.IGNORED,
            }
            notification.update(update_data)
            notification.save()
            msg = ('Notification %(notification_uuid)s of type: %(type)s '
                   'is ignored, because the failover segment is disabled.',
                   {'notification_uuid': notification.notification_uuid,
                    'type': notification.type})
            raise exception.FailoverSegmentDisabled(msg)

        self._process_notification(context, notification)

    @periodic_task.periodic_task(
        spacing=CONF.process_unfinished_notifications_interval)
    def _process_unfinished_notifications(self, context):
        filters = {
            'status': [fields.NotificationStatus.ERROR,
                       fields.NotificationStatus.NEW]
        }
        notifications_list = objects.NotificationList.get_all(context,
                                                              filters=filters)

        for notification in notifications_list:
            if (notification.status == fields.NotificationStatus.ERROR or
                    (notification.status == fields.NotificationStatus.NEW and
                timeutils.is_older_than(
                    notification.generated_time,
                    CONF.retry_notification_new_status_interval))):
                self._process_notification(context, notification)

            # get updated notification from db after workflow execution
            notification_db = objects.Notification.get_by_uuid(
                context, notification.notification_uuid)

            if notification_db.status == fields.NotificationStatus.ERROR:
                # update notification status as failed
                notification_status = fields.NotificationStatus.FAILED
                update_data = {
                    'status': notification_status
                }

                notification_db.update(update_data)
                notification_db.save()
                LOG.error(
                    "Periodic task 'process_unfinished_notifications': "
                    "Notification %(notification_uuid)s exits with "
                    "status: %(status)s.",
                    {'notification_uuid': notification.notification_uuid,
                     'status': notification_status})

    @periodic_task.periodic_task(
        spacing=CONF.check_expired_notifications_interval)
    def _check_expired_notifications(self, context):
        filters = {
            'status': [fields.NotificationStatus.RUNNING,
                       fields.NotificationStatus.ERROR,
                       fields.NotificationStatus.NEW]
        }
        notifications_list = objects.NotificationList.get_all(context,
                                                              filters=filters)

        for notification in notifications_list:
            if timeutils.is_older_than(
                    notification.generated_time,
                    CONF.notifications_expired_interval):
                # update running expired notification status as failed
                notification_status = fields.NotificationStatus.FAILED
                update_data = {
                    'status': notification_status
                }

                notification.update(update_data)
                notification.save()
                LOG.error(
                    "Periodic task 'check_expired_notifications': "
                    "Notification %(notification_uuid)s is expired.",
                    {'notification_uuid': notification.notification_uuid})

    def get_notification_recovery_workflow_details(self, context,
                                                   notification):
        """Retrieve recovery workflow details of the notification"""
        try:
            host_obj = objects.Host.get_by_uuid(
                context, notification.source_host_uuid)
            recovery_method = host_obj.failover_segment.recovery_method

            progress_details = (
                self.driver.get_notification_recovery_workflow_details(
                    context, recovery_method, notification))
            notification['recovery_workflow_details'] = progress_details
        except Exception:
            msg = (_('Failed to fetch notification recovery workflow details '
                     'for %s') % notification.notification_uuid)
            LOG.exception(msg)
            raise exception.MasakariException(msg)

        return notification



# Copyright 2016 NTT Data.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Masakari common internal object model"""

import datetime

from oslo_utils import versionutils
from oslo_versionedobjects import base as ovoo_base
from oslo_versionedobjects import fields as obj_fields

from masakari import objects


def get_attrname(name):
    """Return the mangled name of the attribute's underlying storage."""
    return '_obj_' + name


class MasakariObjectRegistry(ovoo_base.VersionedObjectRegistry):
    notification_classes = []

    def registration_hook(self, cls, index):
        # NOTE(Dinesh_Bhor): This is called when an object is registered,
        # and is responsible for maintaining masakari.objects.$OBJECT
        # as the highest-versioned implementation of a given object.
        version = versionutils.convert_version_to_tuple(cls.VERSION)
        if not hasattr(objects, cls.obj_name()):
            setattr(objects, cls.obj_name(), cls)
        else:
            cur_version = versionutils.convert_version_to_tuple(
                getattr(objects, cls.obj_name()).VERSION)
            if version >= cur_version:
                setattr(objects, cls.obj_name(), cls)

    @classmethod
    def register_notification(cls, notification_cls):
        """Register a class as notification.

        Use only to register concrete notification or payload classes,
        do not register base classes intended for inheritance only.
        """
        cls.register_if(False)(notification_cls)
        cls.notification_classes.append(notification_cls)
        return notification_cls

    @classmethod
    def register_notification_objects(cls):
        """Register previously decorated notification as normal ovos.

        This is not intended for production use but only for testing and
        document generation purposes.
        """
        for notification_cls in cls.notification_classes:
            cls.register(notification_cls)


remotable_classmethod = ovoo_base.remotable_classmethod
remotable = ovoo_base.remotable


class MasakariObject(ovoo_base.VersionedObject):
    """Base class and object factory.

    This forms the base of all objects that can be remoted or instantiated
    via RPC. Simply defining a class that inherits from this base class
    will make it remotely instantiatable. Objects should implement the
    necessary "get" classmethod routines as well as "save" object methods
    as appropriate.
    """

    OBJ_SERIAL_NAMESPACE = 'masakari_object'
    OBJ_PROJECT_NAMESPACE = 'masakari'

    def masakari_obj_get_changes(self):
        """Returns a dict of changed fields with tz unaware datetimes.

        Any timezone aware datetime field will be converted to UTC timezone
        and returned as timezone unaware datetime.

        This will allow us to pass these fields directly to a db update
        method as they can't have timezone information.
        """
        # Get dirtied/changed fields
        changes = self.obj_get_changes()

        # Look for datetime objects that contain timezone information
        for k, v in changes.items():
            if isinstance(v, datetime.datetime) and v.tzinfo:
                # Remove timezone information and adjust the time according to
                # the timezone information's offset.
                changes[k] = v.replace(tzinfo=None) - v.utcoffset()

        # Return modified dict
        return changes

    def obj_reset_changes(self, fields=None, recursive=False):
        """Reset the list of fields that have been changed.

        .. note::

          - This is NOT "revert to previous values"
          - Specifying fields on recursive resets will only be honored at the
            top level. Everything below the top will reset all.

        :param fields: List of fields to reset, or "all" if None.
        :param recursive: Call obj_reset_changes(recursive=True) on
                          any sub-objects within the list of fields
                          being reset.
        """
        if recursive:
            for field in self.obj_get_changes():

                # Ignore fields not in requested set (if applicable)
                if fields and field not in fields:
                    continue

                # Skip any fields that are unset
                if not self.obj_attr_is_set(field):
                    continue

                value = getattr(self, field)

                # Don't reset nulled fields
                if value is None:
                    continue

                # Reset straight Object and ListOfObjects fields
                if isinstance(self.fields[field], obj_fields.ObjectField):
                    value.obj_reset_changes(recursive=True)
                elif isinstance(self.fields[field],
                                obj_fields.ListOfObjectsField):
                    for thing in value:
                        thing.obj_reset_changes(recursive=True)

        if fields:
            self._changed_fields -= set(fields)
        else:
            self._changed_fields.clear()


class MasakariObjectDictCompat(ovoo_base.VersionedObjectDictCompat):
    def __iter__(self):
        for name in self.obj_fields:
            if (self.obj_attr_is_set(name) or
                    name in self.obj_extra_fields):
                yield name

    def keys(self):
        return list(self)


class MasakariTimestampObject(object):
    """Mixin class for db backed objects with timestamp fields.

    Sqlalchemy models that inherit from the oslo_db TimestampMixin will include
    these fields and the corresponding objects will benefit from this mixin.
    """
    fields = {
        'created_at': obj_fields.DateTimeField(nullable=True),
        'updated_at': obj_fields.DateTimeField(nullable=True),
        }


class MasakariPersistentObject(object):
    """Mixin class for Persistent objects.

    This adds the fields that we use in common for most persistent objects.
    """
    fields = {
        'created_at': obj_fields.DateTimeField(nullable=True),
        'updated_at': obj_fields.DateTimeField(nullable=True),
        'deleted_at': obj_fields.DateTimeField(nullable=True),
        'deleted': obj_fields.BooleanField(default=False),
        }


class ObjectListBase(ovoo_base.ObjectListBase):

    @classmethod
    def _obj_primitive_key(cls, field):
        return 'masakari_object.%s' % field

    @classmethod
    def _obj_primitive_field(cls, primitive, field,
                             default=obj_fields.UnspecifiedDefault):
        key = cls._obj_primitive_key(field)
        if default == obj_fields.UnspecifiedDefault:
            return primitive[key]
        else:
            return primitive.get(key, default)


class MasakariObjectSerializer(ovoo_base.VersionedObjectSerializer):
    """A Masakari Object Serializer.

       This implements the Oslo Serializer interface and provides
       the ability to serialize and deserialize MasakariObject entities.
       Any service that needs to accept or return MasakariObjects
       as arguments or result values should pass this to its RPCClient
       and RPCServer objects.

    """
    OBJ_BASE_CLASS = MasakariObject

    def __init__(self):
        super(MasakariObjectSerializer, self).__init__()


def obj_make_list(context, list_obj, item_cls, db_list, **extra_args):
    """Construct an object list from a list of primitives.

    This calls item_cls._from_db_object() on each item of db_list, and
    adds the resulting object to list_obj.

    :param:context: Request context
    :param:list_obj: An ObjectListBase object
    :param:item_cls: The MasakariObject class of the objects within the list
    :param:db_list: The list of primitives to convert to objects
    :param:extra_args: Extra arguments to pass to _from_db_object()
    :returns: list_obj
    """
    list_obj.objects = []
    for db_item in db_list:
        item = item_cls._from_db_object(context, item_cls(), db_item,
                                        **extra_args)
        list_obj.objects.append(item)
    list_obj._context = context
    list_obj.obj_reset_changes()
    return list_obj


def obj_equal_prims(obj_1, obj_2, ignore=None):
    """Compare two primitives for equivalence ignoring some keys.

    This operation tests the primitives of two objects for equivalence.
    Object primitives may contain a list identifying fields that have been
    changed - this is ignored in the comparison. The ignore parameter lists
    any other keys to be ignored.

    :param:obj1: The first object in the comparison
    :param:obj2: The second object in the comparison
    :param:ignore: A list of fields to ignore
    :returns: True if the primitives are equal ignoring changes
    and specified fields, otherwise False.
    """

    def _strip(prim, keys):
        if isinstance(prim, dict):
            for k in keys:
                prim.pop(k, None)
            for v in prim.values():
                _strip(v, keys)
        if isinstance(prim, list):
            for v in prim:
                _strip(v, keys)
        return prim

    if ignore is not None:
        keys = ['masakari_object.changes'] + ignore
    else:
        keys = ['masakari_object.changes']
    prim_1 = _strip(obj_1.obj_to_primitive(), keys)
    prim_2 = _strip(obj_2.obj_to_primitive(), keys)
    return prim_1 == prim_2



# Copyright 2016 NTT Data.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

from alembic import command as alembic_api
from alembic import config as alembic_config
from alembic import migration as alembic_migration
from oslo_config import cfg
from oslo_db import options
from oslo_log import log as logging
import sqlalchemy as sa

import masakari.conf
from masakari.db import api as db_api
from masakari.engine import driver
from masakari import exception
from masakari.i18n import _

options.set_defaults(cfg.CONF)

LOG = logging.getLogger(__name__)
CONF = masakari.conf.CONF


def _migrate_legacy_database(engine, connection, config):
    """Check if database is a legacy sqlalchemy-migrate-managed database.

    If it is, migrate it by "stamping" the initial alembic schema.
    """
    # If the database doesn't have the sqlalchemy-migrate legacy migration
    # table, we don't have anything to do
    if not sa.inspect(engine).has_table('migrate_version'):
        return

    # Likewise, if we've already migrated to alembic, we don't have anything to
    # do
    context = alembic_migration.MigrationContext.configure(
        connection,
        opts={'version_table': 'masakari_alembic_version'})
    if context.get_current_revision():
        return

    # We have legacy migrations but no alembic migration. Stamp (dummy apply)
    # the initial alembic migration(s). There may be one or two to apply
    # depending on what's already applied.

    # Get the currently applied version of the legacy migrations using table
    # reflection to avoid a dependency on sqlalchemy-migrate
    # https://opendev.org/x/sqlalchemy-migrate/src/commit/5d1f322542cd8eb42381612765be4ed9ca8105ec/migrate/versioning/schema.py#L175-L179
    meta = sa.MetaData()
    table = sa.Table('migrate_version', meta, autoload_with=engine)
    with engine.connect() as conn:
        version = conn.execute(sa.select(table.c.version)).scalar()

    # If the user is requesting a skip-level upgrade from a very old version,
    # we can't help them since we don't have alembic-versions of those old
    # migrations :(
    if version < 7:
        reason = _(
            'Your database is at version %03d; we only support upgrading '
            'from version 007 or later. Please upgrade your database using '
            'an earlier release of Masakari and then return here.'
        )
        raise exception.InvalidInput(reason % version)
    elif version > 8:
        if os.getenv('FORCE_MASAKARI_DB_SYNC') is None:
            reason = _(
                'Your database is at version %03d; we do not recognise this '
                'version and it is likely you are carrying out-of-tree '
                'migrations. You can still upgrade but we cannot guarantee '
                'things will work as expected. '
                'If you wish to continue, set the FORCE_MASAKARI_DB_SYNC '
                'environment variable to any value and retry.'
            )
            raise exception.InvalidInput(reason % version)
        else:
            msg = _(
                'Your database is at version %03d; we do not recognise this '
                'version but the FORCE_MASAKARI_DB_SYNC environment variable '
                'is set so we are continuing. Things may break. '
                'You have been warned!',
            )
            LOG.warning(msg, version)

    if version == 7:
        alembic_init_version = '8f848eb45d03'
    else:  # 8 or greater (out-of-tree)
        alembic_init_version = '8bdf5929c5a6'

    LOG.info(
        'The database is still under sqlalchemy-migrate control; '
        'fake applying the initial alembic migration'
    )
    alembic_api.stamp(config, alembic_init_version)


def _find_alembic_conf():
    """Get the project's alembic configuration

    :returns: An instance of ``alembic.config.Config``
    """
    path = os.path.join(
        os.path.abspath(os.path.dirname(__file__)),
        'alembic.ini',
    )
    config = alembic_config.Config(os.path.abspath(path))
    # We don't want to use the logger configuration from the file, which is
    # only really intended for the CLI
    # https://stackoverflow.com/a/42691781/613428
    config.attributes['configure_logger'] = False
    return config


def _upgrade_alembic(engine, config, version):
    # re-use the connection rather than creating a new one
    with engine.begin() as connection:
        config.attributes['connection'] = connection
        _migrate_legacy_database(engine, connection, config)
        alembic_api.upgrade(config, version or 'head')


def db_sync(version=None, engine=None):
    """Migrate the database to `version` or the most recent version."""
    # If the user requested a specific version, check if it's an integer: if
    # so, we're almost certainly in sqlalchemy-migrate land and won't support
    # that
    if version is not None and version.isdigit():
        raise ValueError(
            'You requested an sqlalchemy-migrate database version; this is '
            'no longer supported'
        )

    if engine is None:
        engine = db_api.get_engine()

    config = _find_alembic_conf()

    # Discard the URL encoded in alembic.ini in favour of the URL configured
    # for the engine by the database fixtures, casting from
    # 'sqlalchemy.engine.url.URL' to str in the process. This returns a
    # RFC-1738 quoted URL, which means that a password like "foo@" will be
    # turned into "foo%40". This in turns causes a problem for
    # set_main_option() because that uses ConfigParser.set, which (by design)
    # uses *python* interpolation to write the string out ... where "%" is the
    # special python interpolation character! Avoid this mismatch by quoting
    # all %'s for the set below.
    engine_url = str(engine.url).replace('%', '%%')
    config.set_main_option('sqlalchemy.url', str(engine_url))

    # First upgrade ourselves, followed by Taskflow
    LOG.info('Applying migration(s)')
    _upgrade_alembic(engine, config, version)

    # Get the taskflow driver configured, default is 'taskflow_driver',
    # to load persistence tables to store progress details.
    taskflow_driver = driver.load_masakari_driver(CONF.notification_driver)

    if CONF.taskflow.connection:
        taskflow_driver.upgrade_backend(CONF.taskflow.connection)
    LOG.info('Migration(s) applied')


def db_version():
    """Get database version."""
    engine = db_api.get_engine()
    with engine.connect() as connection:
        m_context = alembic_migration.MigrationContext.configure(
            connection,
            opts={'version_table': 'masakari_alembic_version'}
        )
        return m_context.get_current_revision()



# Copyright(c) 2022 Inspur
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the vmoves api."""

from unittest import mock

import ddt
from webob import exc

from masakari.api.openstack.ha import vmoves
from masakari import exception
from masakari.ha import api as ha_api
from masakari.objects import base as obj_base
from masakari.objects import notification as notification_obj
from masakari.objects import vmove as vmove_obj
from masakari import test
from masakari.tests.unit.api.openstack import fakes
from masakari.tests.unit import fakes as fakes_data
from masakari.tests import uuidsentinel


def _make_vmove_obj(vmove_dict):
    return vmove_obj.VMove(**vmove_dict)


def _make_vmoves_list(vmove_list):
    return vmove_obj.VMove(objects=[
        _make_vmove_obj(a) for a in vmove_list])


@ddt.ddt
class VMoveTestCase(test.TestCase):
    """Test Case for vmove api."""

    bad_request = exception.ValidationError

    def _set_up(self):
        self.controller = vmoves.VMovesController()
        self.req = fakes.HTTPRequest.blank(
            '/v1/notifications/%s/vmoves' % (
                uuidsentinel.fake_notification1),
            use_admin_context=True)
        self.context = self.req.environ['masakari.context']

    def setUp(self):
        super(VMoveTestCase, self).setUp()
        self._set_up()
        self.host_type_notification = fakes_data.create_fake_notification(
            id=1,
            type="COMPUTE_HOST",
            source_host_uuid=uuidsentinel.fake_host_1,
            status="running",
            notification_uuid=uuidsentinel.fake_host_type_notification,
            payload={'event': 'STOPPED',
                     'host_status': 'NORMAL',
                     'cluster_status': 'ONLINE'}
        )
        self.vm_type_notification = fakes_data.create_fake_notification(
            id=1,
            type="VM",
            source_host_uuid=uuidsentinel.fake_host_2,
            status="running",
            notification_uuid=uuidsentinel.fake_vm_type_notification,
            payload={'event': 'STOPPED',
                     'host_status': 'NORMAL',
                     'cluster_status': 'ONLINE'}
        )
        self.vmove_1 = fakes_data.create_fake_vmove(
            id=1,
            uuid=uuidsentinel.fake_vmove_1,
            notification_uuid=self.host_type_notification.notification_uuid,
            instance_uuid=uuidsentinel.fake_instance_1,
            instance_name='vm-1',
            source_host='node01',
            dest_host='node02',
            start_time='2022-11-22 14:50:22',
            end_time="2022-11-22 14:50:35",
            type="evacuation",
            status='succeeded',
            message=None
        )
        self.vmove_2 = fakes_data.create_fake_vmove(
            id=1,
            uuid=uuidsentinel.fake_vmove_1,
            notification_uuid=self.host_type_notification.notification_uuid,
            instance_uuid=uuidsentinel.fake_instance_1,
            instance_name='vm-1',
            source_host='node01',
            dest_host='node02',
            start_time="2022-11-22 14:50:23",
            end_time="2022-11-22 14:50:38",
            type="evacuation",
            status='succeeded',
            message=None
        )
        self.vmove_list = [self.vmove_1, self.vmove_2]
        self.vmove_list_obj = _make_vmoves_list(self.vmove_list)

    @property
    def app(self):
        return fakes.wsgi_app_v1(init_only='vmoves')

    def _assert_vmove_data(self, expected, actual):
        self.assertTrue(obj_base.obj_equal_prims(expected, actual),
                        "The vmove objects were not equal")

    @mock.patch.object(notification_obj.Notification, 'get_by_uuid')
    @mock.patch.object(ha_api.VMoveAPI, 'get_all')
    def test_index(self, mock_get_all, mock_notification):
        mock_notification.return_value = mock.Mock()
        mock_get_all.return_value = self.vmove_list

        result = self.controller.index(
            self.req, uuidsentinel.fake_host_type_notification)
        result = result['vmoves']
        self._assert_vmove_data(self.vmove_list_obj,
                                _make_vmoves_list(result))

    @ddt.data('sort_key', 'sort_dir')
    @mock.patch.object(notification_obj.Notification, 'get_by_uuid',
                       return_value=mock.Mock())
    def test_index_invalid(self, sort_by, mock_notification):
        req = fakes.HTTPRequest.blank(
            '/v1/notifications/%s/vmoves?%s=abcd' % (
                uuidsentinel.fake_notification, sort_by),
            use_admin_context=True)
        self.assertRaises(exc.HTTPBadRequest, self.controller.index, req,
                          uuidsentinel.fake_notification1)

    @mock.patch.object(notification_obj.Notification, 'get_by_uuid')
    @mock.patch.object(ha_api.VMoveAPI, 'get_all')
    def test_index_with_valid_notification(self, mock_get_all,
            mock_notification):
        mock_notification.return_value = mock.Mock()
        mock_get_all.side_effect = exception.NotificationWithoutVMoves
        req = fakes.HTTPRequest.blank('/v1/notifications/%s/vmoves' % (
            uuidsentinel.fake_vm_type_notification), use_admin_context=True)
        self.assertRaises(exc.HTTPBadRequest, self.controller.index, req,
                          uuidsentinel.fake_notification1)

    @mock.patch.object(ha_api.VMoveAPI, 'get_vmove')
    def test_show(self, mock_get_vmove):
        mock_get_vmove.return_value = self.vmove_1
        result = self.controller.show(self.req,
                                      uuidsentinel.fake_notification1,
                                      uuidsentinel.fake_vmove_1)
        vmove = result['vmove']
        self._assert_vmove_data(self.vmove_1,
                                _make_vmove_obj(vmove))

    @mock.patch.object(ha_api.VMoveAPI, 'get_vmove')
    def test_show_with_non_existing_id(self, mock_get_vmove):
        mock_get_vmove.side_effect = exception.VMoveNotFound(id="2")
        self.assertRaises(exc.HTTPNotFound,
                          self.controller.show, self.req,
                          uuidsentinel.fake_notification1, "2")


class VMoveTestCasePolicyNotAuthorized(test.NoDBTestCase):
    """Test Case for vmove non admin."""

    def _set_up(self):
        self.controller = vmoves.VMovesController()
        self.req = fakes.HTTPRequest.blank(
            '/v1/notifications/%s/vmoves' % (
                uuidsentinel.fake_notification1))
        self.context = self.req.environ['masakari.context']

    def setUp(self):
        super(VMoveTestCasePolicyNotAuthorized, self).setUp()
        self._set_up()

    def _check_rule(self, exc, rule_name):
        self.assertEqual(
            "Policy doesn't allow %s to be performed." % rule_name,
            exc.format_message())

    def test_index_no_admin(self):
        rule_name = "os_masakari_api:vmoves:index"
        self.policy.set_rules({rule_name: "project:non_fake"})
        exc = self.assertRaises(exception.PolicyNotAuthorized,
                                self.controller.index,
                                self.req, uuidsentinel.fake_notification1)
        self._check_rule(exc, rule_name)

    def test_show_no_admin(self):
        rule_name = "os_masakari_api:vmoves:detail"
        self.policy.set_rules({rule_name: "project:non_fake"})
        exc = self.assertRaises(exception.PolicyNotAuthorized,
                                self.controller.show,
                                self.req, uuidsentinel.fake_notification1,
                                uuidsentinel.fake_vmove_1)
        self._check_rule(exc, rule_name)



# Copyright 2016 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Starter script for Masakari Engine."""

import sys

from oslo_log import log as logging

import masakari.conf
from masakari import config
from masakari import objects
from masakari import service
from masakari import utils


CONF = masakari.conf.CONF


def main():
    config.parse_args(sys.argv)
    logging.setup(CONF, "masakari")
    utils.monkey_patch()
    objects.register_all()

    server = service.Service.create(binary='masakari-engine',
                                    topic=CONF.masakari_topic)
    service.serve(server)
    service.wait()



#    Copyright 2016 NTT DATA
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
from unittest import mock

from oslo_utils import timeutils

from masakari.api import utils as api_utils
from masakari import db
from masakari import exception
from masakari.objects import fields
from masakari.objects import host
from masakari.tests.unit import fakes as fakes_data
from masakari.tests.unit.objects import test_objects
from masakari.tests import uuidsentinel

NOW = timeutils.utcnow().replace(microsecond=0)

fake_segment_dict = {
    'name': 'fake_segment',
    'recovery_method': 'auto',
    'description': 'fake',
    'service_type': 'CINDER',
    'enabled': True,
    'id': 123,
    'uuid': uuidsentinel.fake_segment,
    'created_at': NOW,
    'updated_at': None,
    'deleted_at': None,
    'deleted': False
}


def _fake_host(**kwargs):
    fake_host = {
        'created_at': NOW,
        'updated_at': None,
        'deleted_at': None,
        'deleted': False,
        'id': 123,
        'uuid': uuidsentinel.fake_host,
        'name': 'fake-host',
        'reserved': False,
        'on_maintenance': False,
        'control_attributes': 'fake_control_attributes',
        'type': 'SSH',
        'failover_segment': fake_segment_dict
        }
    fake_host.update(kwargs)
    return fake_host


fake_host = _fake_host()


class TestHostObject(test_objects._LocalTest):

    def _compare_segment_and_host_data(self, obj):

        self.compare_obj(obj.failover_segment, fake_segment_dict)
        self.assertEqual(obj.name, fake_host.get('name'))
        self.assertEqual(obj.reserved, fake_host.get('reserved'))
        self.assertEqual(obj.on_maintenance, fake_host.get('on_maintenance'))
        self.assertEqual(obj.type, fake_host.get('type'))
        self.assertEqual(obj.control_attributes, fake_host.get('control_attri'
                                                               'butes'))
        self.assertEqual(obj.id, 123)

    def _test_query(self, db_method, obj_method, *args, **kwargs):
        with mock.patch.object(db, db_method) as mock_db:

            db_exception = kwargs.pop('db_exception', None)
            if db_exception:
                mock_db.side_effect = db_exception
            else:
                mock_db.return_value = fake_host

            obj = getattr(host.Host, obj_method)(self.context, *args, **kwargs)
            if db_exception:
                self.assertIsNone(obj)

            self._compare_segment_and_host_data(obj)

    def test_get_by_id(self):
        self._test_query('host_get_by_id', 'get_by_id', 123)

    def test_get_by_uuid(self):
        self._test_query('host_get_by_uuid', 'get_by_uuid',
                         uuidsentinel.fake_segment)

    def test_get_by_name(self):
        self._test_query('host_get_by_name', 'get_by_name',
                         'fake-host')

    def _host_create_attributes(self):

        failover_segment = fakes_data.create_fake_failover_segment(
            name="fake_segment", id=123, description="fakefake",
            service_type="COMPUTE", recovery_method="auto",
            uuid=uuidsentinel.fake_segment
        )

        host_obj = host.Host(context=self.context)
        host_obj.name = 'foo-host'
        host_obj.failover_segment = failover_segment
        host_obj.type = 'fake-type'
        host_obj.reserved = False
        host_obj.on_maintenance = False
        host_obj.control_attributes = 'fake_attributes'
        host_obj.uuid = uuidsentinel.fake_host

        return host_obj

    @mock.patch.object(api_utils, 'notify_about_host_api')
    @mock.patch.object(db, 'host_create')
    def test_create(self, mock_db_create, mock_notify_about_host_api):

        mock_db_create.return_value = fake_host
        host_obj = self._host_create_attributes()
        host_obj.create()
        self._compare_segment_and_host_data(host_obj)
        mock_db_create.assert_called_once_with(self.context, {
            'failover_segment_id': uuidsentinel.fake_segment,
            'on_maintenance': False, 'uuid': uuidsentinel.fake_host,
            'reserved': False, 'name': 'foo-host',
            'control_attributes': 'fake_attributes',
            'type': 'fake-type'})
        action = fields.EventNotificationAction.HOST_CREATE
        phase_start = fields.EventNotificationPhase.START
        phase_end = fields.EventNotificationPhase.END
        notify_calls = [
            mock.call(self.context, host_obj, action=action,
                      phase=phase_start),
            mock.call(self.context, host_obj, action=action,
                      phase=phase_end)]
        mock_notify_about_host_api.assert_has_calls(notify_calls)

    @mock.patch.object(api_utils, 'notify_about_host_api')
    @mock.patch.object(db, 'host_create')
    def test_recreate_fails(self, mock_host_create,
                            mock_notify_about_host_api):
        mock_host_create.return_value = fake_host
        host_obj = self._host_create_attributes()
        host_obj.create()

        self.assertRaises(exception.ObjectActionError, host_obj.create)
        action = fields.EventNotificationAction.HOST_CREATE
        phase_start = fields.EventNotificationPhase.START
        phase_end = fields.EventNotificationPhase.END
        notify_calls = [
            mock.call(self.context, host_obj, action=action,
                      phase=phase_start),
            mock.call(self.context, host_obj, action=action,
                      phase=phase_end)]
        mock_notify_about_host_api.assert_has_calls(notify_calls)

        mock_host_create.assert_called_once_with(self.context, {
            'uuid': uuidsentinel.fake_host, 'name': 'foo-host',
            'failover_segment_id': uuidsentinel.fake_segment,
            'type': 'fake-type', 'reserved': False, 'on_maintenance': False,
            'control_attributes': 'fake_attributes'})

    @mock.patch.object(api_utils, 'notify_about_host_api')
    @mock.patch.object(db, 'host_delete')
    def test_destroy(self, mock_host_destroy, mock_notify_about_host_api):
        host_obj = self._host_create_attributes()
        host_obj.id = 123
        host_obj.destroy()

        mock_host_destroy.assert_called_once_with(
            self.context, uuidsentinel.fake_host)
        action = fields.EventNotificationAction.HOST_DELETE
        phase_start = fields.EventNotificationPhase.START
        phase_end = fields.EventNotificationPhase.END
        notify_calls = [
            mock.call(self.context, host_obj, action=action,
                      phase=phase_start),
            mock.call(self.context, host_obj, action=action,
                      phase=phase_end)]
        mock_notify_about_host_api.assert_has_calls(notify_calls)

    @mock.patch.object(api_utils, 'notify_about_host_api')
    @mock.patch.object(db, 'host_delete')
    def test_destroy_host_not_found(self, mock_host_destroy,
                                    mock_notify_about_host_api):
        mock_host_destroy.side_effect = exception.HostNotFound(id=123)
        host_obj = self._host_create_attributes()
        host_obj.id = 123
        self.assertRaises(exception.HostNotFound, host_obj.destroy)
        action = fields.EventNotificationAction.HOST_DELETE
        phase_start = fields.EventNotificationPhase.START
        notify_calls = [
            mock.call(self.context, host_obj, action=action,
                      phase=phase_start)]
        mock_notify_about_host_api.assert_has_calls(notify_calls)

    @mock.patch.object(db, 'host_get_all_by_filters')
    def test_get_host_by_filters(self, mock_api_get):
        fake_host2 = copy.deepcopy(fake_host)
        fake_host2['name'] = 'fake_host22'

        mock_api_get.return_value = [fake_host2, fake_host]

        filters = {'reserved': False}
        host_result = host.HostList.get_all(self.context, filters=filters)
        self.assertEqual(2, len(host_result))
        mock_api_get.assert_called_once_with(self.context, filters={
            'reserved': False
        }, limit=None, marker=None, sort_dirs=None, sort_keys=None)

    @mock.patch.object(db, 'host_get_all_by_filters')
    def test_get_limit_and_marker_invalid_marker(self, mock_api_get):
        host_name = 'fake-host'
        mock_api_get.side_effect = exception.MarkerNotFound(marker=host_name)

        self.assertRaises(exception.MarkerNotFound,
                          host.HostList.get_all,
                          self.context, limit=5, marker=host_name)

    @mock.patch.object(api_utils, 'notify_about_host_api')
    @mock.patch('masakari.objects.base.MasakariObject'
                '.masakari_obj_get_changes')
    @mock.patch.object(db, 'host_update')
    def test_save(
            self, mock_host_update, mock_masakari_obj_get_changes,
            mock_notify_about_host_api):

        mock_host_update.return_value = fake_host

        host_data = {
            'name': 'foo-host',
            "uuid": uuidsentinel.fake_host,
            'id': 123
        }

        mock_masakari_obj_get_changes.return_value = host_data
        host_obj = self._host_create_attributes()
        host_obj.id = 123
        host_obj.save()
        self._compare_segment_and_host_data(host_obj)
        (mock_host_update.
         assert_called_once_with(self.context, uuidsentinel.fake_host,
                                 {'name': 'foo-host',
                                  'uuid': uuidsentinel.fake_host
                                  }))
        action = fields.EventNotificationAction.HOST_UPDATE
        phase_start = fields.EventNotificationPhase.START
        phase_end = fields.EventNotificationPhase.END
        notify_calls = [
            mock.call(self.context, host_obj, action=action,
                      phase=phase_start),
            mock.call(self.context, host_obj, action=action,
                      phase=phase_end)]
        mock_notify_about_host_api.assert_has_calls(notify_calls)

    @mock.patch.object(db, 'host_update')
    def test_save_lazy_attribute_changed(self, mock_host_update):

        mock_host_update.return_value = fake_host

        host_obj = self._host_create_attributes()
        host_obj.id = 123
        self.assertRaises(exception.ObjectActionError, host_obj.save)

    @mock.patch.object(api_utils, 'notify_about_host_api')
    @mock.patch.object(db, 'host_update')
    def test_save_host_already_exists(self, mock_host_update,
                                      mock_notify_about_host_api):

        mock_host_update.side_effect = exception.HostExists(name="foo-host")

        host_object = host.Host(context=self.context)
        host_object.name = "foo-host"
        host_object.id = 123
        host_object.uuid = uuidsentinel.fake_host

        self.assertRaises(exception.HostExists, host_object.save)
        action = fields.EventNotificationAction.HOST_UPDATE
        phase_start = fields.EventNotificationPhase.START
        notify_calls = [
            mock.call(self.context, host_object, action=action,
                      phase=phase_start)]
        mock_notify_about_host_api.assert_has_calls(notify_calls)

    @mock.patch.object(api_utils, 'notify_about_host_api')
    @mock.patch.object(db, 'host_update')
    def test_save_host_not_found(self, mock_host_update,
                                 mock_notify_about_host_api):

        mock_host_update.side_effect = exception.HostNotFound(id="foo-host")

        host_object = host.Host(context=self.context)
        host_object.name = "foo-host"
        host_object.id = 123
        host_object.uuid = uuidsentinel.fake_host

        self.assertRaises(exception.HostNotFound, host_object.save)
        action = fields.EventNotificationAction.HOST_UPDATE
        phase_start = fields.EventNotificationPhase.START
        notify_calls = [
            mock.call(self.context, host_object, action=action,
                      phase=phase_start)]
        mock_notify_about_host_api.assert_has_calls(notify_calls)

    def _host_create_attributes2(self):

        failover_segment = fakes_data.create_fake_failover_segment(
            name="fake_segment", id=123, description="fakefake",
            service_type="COMPUTE", recovery_method="auto",
            uuid=uuidsentinel.fake_segment
        )

        host_obj = self._host_create_attributes()
        host_obj.failover_segment_id = failover_segment.uuid

        return host_obj

    def test_obj_make_compatible(self):
        host_obj = self._host_create_attributes2()
        primitive = host_obj.obj_to_primitive()
        self.assertIn('failover_segment', primitive['masakari_object.data'])
        self.assertNotIn(
            'failover_segment_id', primitive['masakari_object.data'])



# Copyright 2016 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Utilities and helper functions."""

import contextlib
import functools
import inspect
import pyclbr
import shutil
import sys
import tempfile

import eventlet
from oslo_concurrency import lockutils
from oslo_context import context as common_context
from oslo_log import log as logging
from oslo_utils import importutils
from oslo_utils import strutils
from oslo_utils import timeutils

import masakari.conf
from masakari import exception
from masakari.i18n import _
from masakari import safe_utils


CONF = masakari.conf.CONF

LOG = logging.getLogger(__name__)


def reraise(tp, value, tb=None):
    try:
        if value is None:
            value = tp()
        if value.__traceback__ is not tb:
            raise value.with_traceback(tb)
        raise value
    finally:
        value = None
        tb = None


def utf8(value):
    """Try to turn a string into utf-8 if possible.

    The original code was copied from the utf8 function in
    http://github.com/facebook/tornado/blob/master/tornado/escape.py

    """
    if value is None or isinstance(value, bytes):
        return value

    if not isinstance(value, str):
        value = str(value)

    return value.encode('utf-8')


def check_isinstance(obj, cls):
    """Checks that obj is of type cls, and lets PyLint infer types."""
    if isinstance(obj, cls):
        return obj
    raise Exception(_('Expected object of type: %s') % (str(cls)))


def monkey_patch():
    """If the CONF.monkey_patch set as True,
    this function patches a decorator
    for all functions in specified modules.
    You can set decorators for each modules
    using CONF.monkey_patch_modules.
    The format is "Module path:Decorator function".

    name - name of the function
    function - object of the function
    """
    # If CONF.monkey_patch is not True, this function do nothing.
    if not CONF.monkey_patch:
        return

    def is_method(obj):
        # Unbound methods became regular functions on Python 3
        return inspect.ismethod(obj) or inspect.isfunction(obj)

    # Get list of modules and decorators
    for module_and_decorator in CONF.monkey_patch_modules:
        module, decorator_name = module_and_decorator.split(':')
        # import decorator function
        decorator = importutils.import_class(decorator_name)
        __import__(module)
        # Retrieve module information using pyclbr
        module_data = pyclbr.readmodule_ex(module)
        for key, value in module_data.items():
            # set the decorator for the class methods
            if isinstance(value, pyclbr.Class):
                clz = importutils.import_class("%s.%s" % (module, key))
                for method, func in inspect.getmembers(clz, is_method):
                    setattr(clz, method,
                            decorator("%s.%s.%s" % (module, key,
                                                    method), func))
            # set the decorator for the function
            if isinstance(value, pyclbr.Function):
                func = importutils.import_class("%s.%s" % (module, key))
                setattr(sys.modules[module], key,
                        decorator("%s.%s" % (module, key), func))


def walk_class_hierarchy(clazz, encountered=None):
    """Walk class hierarchy, yielding most derived classes first."""
    if not encountered:
        encountered = []
    for subclass in clazz.__subclasses__():
        if subclass not in encountered:
            encountered.append(subclass)
            # drill down to leaves first
            for subsubclass in walk_class_hierarchy(subclass, encountered):
                yield subsubclass
            yield subclass


def expects_func_args(*args):
    def _decorator_checker(dec):
        @functools.wraps(dec)
        def _decorator(f):
            base_f = safe_utils.get_wrapped_function(f)
            arg_names, a, kw, _, _, _, _ = inspect.getfullargspec(base_f)
            if a or kw or set(args) <= set(arg_names):
                # NOTE : We can't really tell if correct stuff will
                # be passed if it's a function with *args or **kwargs so
                # we still carry on and hope for the best
                return dec(f)
            else:
                raise TypeError("Decorated function %(f_name)s does not "
                                "have the arguments expected by the "
                                "decorator %(d_name)s" %
                                {'f_name': base_f.__name__,
                                 'd_name': dec.__name__})
        return _decorator
    return _decorator_checker


def isotime(at=None):
    """Current time as ISO string,
    as timeutils.isotime() is deprecated

    :returns: Current time in ISO format
    """
    if not at:
        at = timeutils.utcnow()
    date_string = at.strftime("%Y-%m-%dT%H:%M:%S")
    tz = at.tzinfo.tzname(None) if at.tzinfo else 'UTC'
    date_string += ('Z' if tz in ['UTC', 'UTC+00:00'] else tz)
    return date_string


def strtime(at):
    return at.strftime("%Y-%m-%dT%H:%M:%S.%f")


class ExceptionHelper(object):
    """Class to wrap another and translate the ClientExceptions raised by its
    function calls to the actual ones.
    """

    def __init__(self, target):
        self._target = target

    def __getattr__(self, name):
        func = getattr(self._target, name)

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as e:
                reraise(*e.exc_info)
        return wrapper


def spawn(func, *args, **kwargs):
    """Passthrough method for eventlet.spawn.

    This utility exists so that it can be stubbed for testing without
    interfering with the service spawns.

    It will also grab the context from the threadlocal store and add it to
    the store on the new thread.  This allows for continuity in logging the
    context when using this method to spawn a new thread.
    """
    _context = common_context.get_current()

    @functools.wraps(func)
    def context_wrapper(*args, **kwargs):
        # NOTE: If update_store is not called after spawn it won't be
        # available for the logger to pull from threadlocal storage.
        if _context is not None:
            _context.update_store()
        return func(*args, **kwargs)

    return eventlet.spawn(context_wrapper, *args, **kwargs)


def spawn_n(func, *args, **kwargs):
    """Passthrough method for eventlet.spawn_n.

    This utility exists so that it can be stubbed for testing without
    interfering with the service spawns.

    It will also grab the context from the threadlocal store and add it to
    the store on the new thread.  This allows for continuity in logging the
    context when using this method to spawn a new thread.
    """
    _context = common_context.get_current()

    @functools.wraps(func)
    def context_wrapper(*args, **kwargs):
        # NOTE: If update_store is not called after spawn_n it won't be
        # available for the logger to pull from threadlocal storage.
        if _context is not None:
            _context.update_store()
        func(*args, **kwargs)

    eventlet.spawn_n(context_wrapper, *args, **kwargs)


@contextlib.contextmanager
def tempdir(**kwargs):
    argdict = kwargs.copy()
    if 'dir' not in argdict:
        argdict['dir'] = CONF.tempdir
    tmpdir = tempfile.mkdtemp(**argdict)
    try:
        yield tmpdir
    finally:
        try:
            shutil.rmtree(tmpdir)
        except OSError as e:
            LOG.error('Could not remove tmpdir: %s', e)


def validate_integer(value, name, min_value=None, max_value=None):
    """Make sure that value is a valid integer, potentially within range."""
    try:
        return strutils.validate_integer(value, name, min_value, max_value)
    except ValueError as e:
        raise exception.InvalidInput(reason=e)


def synchronized(name, semaphores=None, blocking=False):
    def wrap(f):
        @functools.wraps(f)
        def inner(*args, **kwargs):
            lock_name = 'masakari-%s' % name
            int_lock = lockutils.internal_lock(lock_name,
                                               semaphores=semaphores)
            LOG.debug("Acquiring lock: %(lock_name)s on resource: "
                      "%(resource)s", {'lock_name': lock_name,
                                       'resource': f.__name__})

            if not int_lock.acquire(blocking=blocking):
                raise exception.LockAlreadyAcquired(resource=name)
            try:
                return f(*args, **kwargs)
            finally:
                LOG.debug("Releasing lock: %(lock_name)s on resource: "
                          "%(resource)s", {'lock_name': lock_name,
                                           'resource': f.__name__})
                int_lock.release()
        return inner
    return wrap



# Copyright (c) 2020 NTT Data
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from masakari.api.openstack import common


def get_view_builder(req):
    base_url = req.application_url
    return ViewBuilder(base_url)


class ViewBuilder(common.ViewBuilder):

    def __init__(self, base_url):
        """:param base_url: url of the root wsgi application."""
        self.prefix = self._update_masakari_link_prefix(base_url)
        self.base_url = base_url

    def _host_details(self, host):
        return {
            'id': host.id,
            'uuid': host.uuid,
            'name': host.name,
            'failover_segment_id': host.failover_segment.uuid,
            'failover_segment': host.failover_segment,
            'type': host.type,
            'reserved': host.reserved,
            'control_attributes': host.control_attributes,
            'on_maintenance': host.on_maintenance,
            'created_at': host.created_at,
            'updated_at': host.updated_at,
            'deleted_at': host.deleted_at,
            'deleted': host.deleted
        }

    def build_host(self, host):
        get_host_response = self._host_details(host)
        return get_host_response

    def build_hosts(self, hosts):
        host_objs = []
        for host in hosts:
            get_host_response = self._host_details(host)
            host_objs.append(get_host_response)
        return dict(hosts=host_objs)



# Copyright 2016 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
RequestContext: context for requests that persist through all of masakari.
"""

import copy

from keystoneauth1.access import service_catalog as ksa_service_catalog
from keystoneauth1 import plugin
from oslo_context import context
from oslo_db.sqlalchemy import enginefacade
from oslo_log import log as logging
from oslo_utils import timeutils

from masakari import exception
from masakari.i18n import _
from masakari import policy
from masakari import utils

LOG = logging.getLogger(__name__)


class _ContextAuthPlugin(plugin.BaseAuthPlugin):
    """A keystoneauth auth plugin that uses the values from the Context.

    Ideally we would use the plugin provided by auth_token middleware however
    this plugin isn't serialized yet so we construct one from the serialized
    auth data.
    """

    def __init__(self, auth_token, sc):
        super(_ContextAuthPlugin, self).__init__()

        self.auth_token = auth_token
        self.service_catalog = ksa_service_catalog.ServiceCatalogV2(sc)

    def get_token(self, *args, **kwargs):
        return self.auth_token

    def get_endpoint(self, session, service_type=None, interface=None,
                     region_name=None, service_name=None, **kwargs):
        return self.service_catalog.url_for(service_type=service_type,
                                            service_name=service_name,
                                            interface=interface,
                                            region_name=region_name)


@enginefacade.transaction_context_provider
class RequestContext(context.RequestContext):
    """Security context and request information.

    Represents the user taking a given action within the system.

    """
    FROM_DICT_EXTRA_KEYS = [
        'read_deleted', 'remote_address',
        'timestamp', 'service_catalog',
    ]

    def __init__(self, user_id=None, project_id=None,
                 is_admin=None, read_deleted="no",
                 roles=None, remote_address=None, timestamp=None,
                 request_id=None, auth_token=None, overwrite=True,
                 user_name=None, project_name=None, service_catalog=None,
                 user_auth_plugin=None, **kwargs):
        """:param read_deleted: 'no' indicates deleted records are hidden,
                'yes' indicates deleted records are visible,
                'only' indicates that *only* deleted records are visible.

           :param overwrite: Set to False to ensure that the greenthread local
                copy of the index is not overwritten.

           :param user_auth_plugin: The auth plugin for the current request's
                authentication data.

           :param kwargs: Extra arguments that might be present, but we ignore
                because they possibly came in from older rpc messages.
        """
        user = kwargs.pop('user', None)
        tenant = kwargs.pop('tenant', None)
        super(RequestContext, self).__init__(
            auth_token=auth_token,
            user_id=user_id or user,
            project_id=project_id or tenant,
            domain_id=kwargs.pop('domain', None),
            user_domain_id=kwargs.pop('user_domain', None),
            project_domain_id=kwargs.pop('project_domain', None),
            is_admin=is_admin,
            read_only=kwargs.pop('read_only', False),
            show_deleted=kwargs.pop('show_deleted', False),
            request_id=request_id,
            resource_uuid=kwargs.pop('resource_uuid', None),
            overwrite=overwrite,
            roles=roles,
            user_name=user_name,
            project_name=project_name,
            is_admin_project=kwargs.pop('is_admin_project', True),
            global_request_id=kwargs.pop('global_request_id', None))
        # oslo_context's RequestContext.to_dict() generates this field, we can
        # safely ignore this as we don't use it.
        kwargs.pop('user_identity', None)
        if kwargs:
            LOG.debug('Arguments dropped when creating context: %s',
                      str(kwargs))

        if read_deleted is None:
            # If we did not get a value for read_deleted, ensure we default
            # it to "no" as code expects it to be a string.
            self.read_deleted = 'no'
        else:
            self.read_deleted = read_deleted
        self.remote_address = remote_address
        if not timestamp:
            timestamp = timeutils.utcnow()
        if isinstance(timestamp, str):
            timestamp = timeutils.parse_strtime(timestamp)
        self.timestamp = timestamp

        if service_catalog:
            # Only include required parts of service_catalog
            self.service_catalog = [
                s for s in service_catalog if s.get('type') in (
                    'compute', 'identity')]
        else:
            # if list is empty or none
            self.service_catalog = []

        self.user_auth_plugin = user_auth_plugin
        if self.is_admin is None:
            self.is_admin = policy.check_is_admin(self)

    def get_auth_plugin(self):
        if self.user_auth_plugin:
            return self.user_auth_plugin
        else:
            return _ContextAuthPlugin(self.auth_token, self.service_catalog)

    def _get_read_deleted(self):
        return self._read_deleted

    def _set_read_deleted(self, read_deleted):
        if read_deleted not in ('no', 'yes', 'only'):
            raise ValueError(_("read_deleted can only be one of 'no', "
                               "'yes' or 'only', not %r") % read_deleted)
        self._read_deleted = read_deleted

    def _del_read_deleted(self):
        del self._read_deleted

    read_deleted = property(_get_read_deleted, _set_read_deleted,
                            _del_read_deleted)

    def to_dict(self):
        values = super(RequestContext, self).to_dict()
        # FIXME: defensive hasattr() checks need to be
        # removed once we figure out why we are seeing stack
        # traces
        values.update({
            'user_id': getattr(self, 'user_id', None),
            'project_id': getattr(self, 'project_id', None),
            'read_deleted': getattr(self, 'read_deleted', 'no'),
            'remote_address': getattr(self, 'remote_address', None),
            'timestamp': utils.strtime(self.timestamp) if hasattr(
                self, 'timestamp') else None,
            'user_name': getattr(self, 'user_name', None),
            'service_catalog': getattr(self, 'service_catalog', None),
            'project_name': getattr(self, 'project_name', None)
        })
        return values

    def elevated(self, read_deleted=None):
        """Return a version of this context with admin flag set."""
        context = copy.copy(self)
        # context.roles must be deepcopied to leave original roles
        # without changes
        context.roles = copy.deepcopy(self.roles)
        context.is_admin = True

        if 'admin' not in context.roles:
            context.roles.append('admin')

        if read_deleted is not None:
            context.read_deleted = read_deleted

        return context

    def can(self, action, target=None, fatal=True):
        """Verifies that the given action is valid on the target in this
        context.

        :param action: string representing the action to be checked.
        :param target: dictionary representing the object of the action
            for object creation this should be a dictionary representing the
            location of the object e.g. ``{'project_id': context.project_id}``.
            If None, then this default target will be considered:
            {'project_id': self.project_id, 'user_id': self.user_id}
        :param fatal: if False, will return False when an exception.Forbidden
           occurs.

        :raises masakari.exception.Forbidden: if verification fails and fatal
            is True.

        :return: returns a non-False value (not necessarily "True") if
            authorized and False if not authorized and fatal is False.
        """
        if target is None:
            target = {'project_id': self.project_id,
                      'user_id': self.user_id}
        try:
            return policy.authorize(self, action, target)
        except exception.Forbidden:
            if fatal:
                raise
            return False

    def to_policy_values(self):
        policy = super(RequestContext, self).to_policy_values()
        policy['is_admin'] = self.is_admin
        return policy

    def __str__(self):
        return "<Context %s>" % self.to_dict()


def get_admin_context(read_deleted="no"):
    return RequestContext(user_id=None,
                          project_id=None,
                          is_admin=True,
                          read_deleted=read_deleted,
                          overwrite=False)



# Copyright 2016 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

from masakari.conf import api
from masakari.conf import base
from masakari.conf import coordination
from masakari.conf import database
from masakari.conf import engine
from masakari.conf import engine_driver
from masakari.conf import exceptions
from masakari.conf import nova
from masakari.conf import osapi_v1
from masakari.conf import paths
from masakari.conf import service
from masakari.conf import ssl
from masakari.conf import wsgi

CONF = cfg.CONF

api.register_opts(CONF)
base.register_opts(CONF)
coordination.register_opts(CONF)
database.register_opts(CONF)
engine.register_opts(CONF)
engine_driver.register_opts(CONF)
exceptions.register_opts(CONF)
nova.register_opts(CONF)
osapi_v1.register_opts(CONF)
paths.register_opts(CONF)
ssl.register_opts(CONF)
service.register_opts(CONF)
wsgi.register_opts(CONF)



# Copyright (C) 2019 NTT DATA
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import sys

import openstack
from oslotest import base

openstack.enable_logging(
    debug=True,
    http_debug=True,
    stream=sys.stdout,
    format_stream=True,
    format_template='%(asctime)s %(name)-32s %(message)s',
)
#: Defines the OpenStack Client Config (OCC) cloud keys in your OCC config
#: file, typically in /etc/openstack/clouds.yaml. That configuration
#: will determine where the functional tests will be run and what resource
#: defaults will be used to run the functional tests.
ADMIN_CLOUD_NAME = os.getenv(
    "MASAKARI_FUNCTIONAL_TESTS_ADMIN", "devstack-admin"
)
USER_CLOUD_NAME = os.getenv("MASAKARI_FUNCTIONAL_TESTS_USER", "devstack")


class BaseFunctionalTest(base.BaseTestCase):

    def setUp(self, ha_api_version="1.0"):
        super(BaseFunctionalTest, self).setUp()

        self.admin_conn = openstack.connect(
            cloud=ADMIN_CLOUD_NAME,
            ha_api_version=ha_api_version,
        )
        self.conn = openstack.connect(cloud=USER_CLOUD_NAME)
        self.hypervisors = self.admin_conn.list_hypervisors()



# Copyright(c) 2022 Inspur
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The VM Move API extension."""

from http import HTTPStatus
from webob import exc

from masakari.api.openstack import common
from masakari.api.openstack import extensions
from masakari.api.openstack import wsgi
from masakari import exception
from masakari.ha import api as vmove_api
from masakari.policies import vmoves as vmove_policies

ALIAS = "vmoves"


class VMovesController(wsgi.Controller):
    """The VM move API controller for the Instance HA."""

    def __init__(self):
        self.api = vmove_api.VMoveAPI()

    @extensions.expected_errors((HTTPStatus.BAD_REQUEST, HTTPStatus.FORBIDDEN,
                                 HTTPStatus.NOT_FOUND))
    def index(self, req, notification_id):
        """Returns a list of vmoves."""
        context = req.environ['masakari.context']
        context.can(vmove_policies.VMOVES % 'index')

        try:
            filters = {}
            limit, marker = common.get_limit_and_marker(req)
            sort_keys, sort_dirs = common.get_sort_params(req.params)

            if 'status' in req.params:
                filters['status'] = req.params['status']
            if 'type' in req.params:
                filters['type'] = req.params['type']

            vmoves = self.api.get_all(context,
                                      notification_id,
                                      filters=filters,
                                      sort_keys=sort_keys,
                                      sort_dirs=sort_dirs,
                                      limit=limit,
                                      marker=marker)
        except exception.MarkerNotFound as ex:
            raise exc.HTTPBadRequest(explanation=ex.format_message())
        except exception.Invalid as e:
            raise exc.HTTPBadRequest(explanation=e.format_message())
        except exception.NotificationWithoutVMoves as e:
            raise exc.HTTPBadRequest(explanation=e.format_message())
        except exception.NotificationNotFound as ex:
            raise exc.HTTPNotFound(explanation=ex.format_message())

        return {'vmoves': vmoves}

    @extensions.expected_errors((HTTPStatus.BAD_REQUEST, HTTPStatus.FORBIDDEN,
                                 HTTPStatus.NOT_FOUND))
    def show(self, req, notification_id, id):
        """Shows the details of one vmove."""
        context = req.environ['masakari.context']
        context.can(vmove_policies.VMOVES % 'detail')
        try:
            vmove = self.api.get_vmove(context, notification_id, id)
        except exception.NotificationWithoutVMoves as e:
            raise exc.HTTPBadRequest(explanation=e.format_message())
        except exception.VMoveNotFound as e:
            raise exc.HTTPNotFound(explanation=e.format_message())

        return {'vmove': vmove}


class VMoves(extensions.V1APIExtensionBase):
    """vmoves controller"""

    name = "vmoves"
    alias = ALIAS
    version = 1

    def get_resources(self):
        parent = {'member_name': 'notification',
                  'collection_name': 'notifications'}
        resources = [
            extensions.ResourceExtension(
                'vmoves', VMovesController(), parent=parent,
                member_name='vmove')]

        return resources

    def get_controller_extensions(self):
        return []



# Copyright(c) 2022 Inspur
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_policy import policy

from masakari.policies import base


VMOVES = 'os_masakari_api:vmoves:%s'

rules = [
    policy.DocumentedRuleDefault(
        name=VMOVES % 'index',
        check_str=base.RULE_ADMIN_API,
        description="Lists IDs, notification_id, instance_id, source_host, "
                    "dest_host, status and type for all VM moves.",
        operations=[
            {
                'method': 'GET',
                'path': '/notifications/{notification_id}/vmoves'
            }
        ]),
    policy.DocumentedRuleDefault(
        name=VMOVES % 'detail',
        check_str=base.RULE_ADMIN_API,
        description="Shows details for one VM move.",
        operations=[
            {
                'method': 'GET',
                'path': '/notifications/{notification_id}/vmoves/'
                        '{vmove_id}'
            }
        ]),
    policy.RuleDefault(
        name=VMOVES % 'discoverable',
        check_str=base.RULE_ADMIN_API,
        description="VM moves API extensions to change the API.",
        ),
]


def list_rules():
    return rules
